@@ -30,6 +30,7 @@
 #include "ica_api.h"
 #include "icastats.h"
 #include "fips.h"
+#include "rng.h"
 #include "s390_rsa.h"
 #include "s390_ecc.h"
 #include "s390_crypto.h"
@@ -59,6 +60,20 @@ void ica_set_fallback_mode(int fallback_mode)
 		ica_fallbacks_enabled = 0;
 }
 
+int ica_offload_enabled = 0;
+
+void ica_set_offload_mode(int offload_mode)
+{
+	ica_offload_enabled = offload_mode ? 1 : 0;
+}
+
+int ica_stats_enabled = 1;
+
+void ica_set_stats_mode(int stats_mode)
+{
+	ica_stats_enabled = stats_mode ? 1 : 0;
+}
+
 static unsigned int check_des_parms(unsigned int mode,
 				    unsigned long data_length,
 				    const unsigned char *in_data,
@@ -561,6 +576,84 @@ unsigned int ica_sha512(unsigned int message_part,
 			   (uint64_t *) &sha512_context->runningLengthHigh);
 }
 
+unsigned int ica_sha512_224(unsigned int message_part,
+			    uint64_t input_length,
+			    unsigned char *input_data,
+			    sha512_context_t *sha512_context,
+			    unsigned char *output_data)
+{
+	unsigned int rc;
+
+#ifdef ICA_FIPS
+	if (fips >> 1)
+		return EACCES;
+#endif /* ICA_FIPS */
+
+	/* check for obvious errors in parms */
+	if ((input_data == NULL) ||
+	    (sha512_context == NULL) ||
+	    (output_data == NULL))
+		return EINVAL;
+
+	/* make sure some message part is specified */
+	rc = check_message_part(message_part);
+	if (rc)
+		return rc;
+
+	/*
+	 * for FIRST or MIDDLE calls the input
+	 * data length must be a multiple of 128 bytes.
+	 */
+	if (input_length & 0x7f &&
+	    (message_part == SHA_MSG_PART_FIRST ||
+	     message_part == SHA_MSG_PART_MIDDLE))
+		return EINVAL;
+
+	return s390_sha512_224((unsigned char *)&sha512_context->sha512Hash,
+			       input_data, input_length, output_data, message_part,
+			       (uint64_t *) &sha512_context->runningLengthLow,
+			       (uint64_t *) &sha512_context->runningLengthHigh);
+}
+
+unsigned int ica_sha512_256(unsigned int message_part,
+			    uint64_t input_length,
+			    unsigned char *input_data,
+			    sha512_context_t *sha512_context,
+			    unsigned char *output_data)
+{
+	unsigned int rc;
+
+#ifdef ICA_FIPS
+	if (fips >> 1)
+		return EACCES;
+#endif /* ICA_FIPS */
+
+	/* check for obvious errors in parms */
+	if ((input_data == NULL) ||
+	    (sha512_context == NULL) ||
+	    (output_data == NULL))
+		return EINVAL;
+
+	/* make sure some message part is specified */
+	rc = check_message_part(message_part);
+	if (rc)
+		return rc;
+
+	/*
+	 * for FIRST or MIDDLE calls the input
+	 * data length must be a multiple of 128 bytes.
+	 */
+	if (input_length & 0x7f &&
+	    (message_part == SHA_MSG_PART_FIRST ||
+	     message_part == SHA_MSG_PART_MIDDLE))
+		return EINVAL;
+
+	return s390_sha512_256((unsigned char *)&sha512_context->sha512Hash,
+			       input_data, input_length, output_data, message_part,
+			       (uint64_t *) &sha512_context->runningLengthLow,
+			       (uint64_t *) &sha512_context->runningLengthHigh);
+}
+
 unsigned int ica_sha3_224(unsigned int message_part,
 			unsigned int input_length,
 			unsigned char *input_data,
@@ -925,7 +1018,11 @@ unsigned int ica_rsa_mod_expo(ica_adapter_handle_t adapter_handle,
 		rc = ica_fallbacks_enabled ?
 			rsa_mod_expo_sw(&rb) : ENODEV;
 	else {
-		rc = ioctl(adapter_handle, ICARSAMODEXPO, &rb);
+		if (any_card_online)
+			rc = ioctl(adapter_handle, ICARSAMODEXPO, &rb);
+		else
+			rc = ENODEV;
+
 		if (!rc)
 			hardware = ALGO_HW;
 		else
@@ -1034,7 +1131,11 @@ unsigned int ica_rsa_crt(ica_adapter_handle_t adapter_handle,
 		rc = ica_fallbacks_enabled ?
 			rsa_crt_sw(&rb) : ENODEV;
 	else {
-		rc = ioctl(adapter_handle, ICARSACRT, &rb);
+		if (any_card_online)
+			rc = ioctl(adapter_handle, ICARSACRT, &rb);
+		else
+			rc = ENODEV;
+
 		if(!rc)
 			hardware = ALGO_HW;
 		else
@@ -1322,6 +1423,534 @@ void ica_ec_key_free(ICA_EC_KEY *key)
 	free(key);
 }
 
+static inline int check_fips(void)
+{
+#ifdef ICA_FIPS
+	return fips >> 1;
+#else
+	return 0;
+#endif
+}
+
+
+int ica_x25519_ctx_new(ICA_X25519_CTX **ctx)
+{
+	if (!msa9_switch || ctx == NULL)
+		return -1;
+
+	*ctx = calloc(1, sizeof(**ctx));
+	return 0;
+}
+
+int ica_x448_ctx_new(ICA_X448_CTX **ctx)
+{
+	if (!msa9_switch || ctx == NULL)
+		return -1;
+
+	*ctx = calloc(1, sizeof(**ctx));
+	return 0;
+}
+
+int ica_ed25519_ctx_new(ICA_ED25519_CTX **ctx)
+{
+	if (!msa9_switch || ctx == NULL)
+		return -1;
+
+	*ctx = calloc(1, sizeof(**ctx));
+	return 0;
+}
+
+int ica_ed448_ctx_new(ICA_ED448_CTX **ctx)
+{
+	if (!msa9_switch || ctx == NULL)
+		return -1;
+
+	*ctx = calloc(1, sizeof(**ctx));
+	return 0;
+}
+
+int ica_x25519_key_set(ICA_X25519_CTX *ctx,
+		       const unsigned char priv[32],
+		       const unsigned char pub[32])
+{
+	if (check_fips() || !msa9_switch || ctx == NULL)
+		return -1;
+
+	if (priv != NULL) {
+		memcpy(ctx->priv, priv, 32);
+		ctx->priv_init = 1;
+		ctx->pub_init = 0;
+	}
+
+	if (pub != NULL) {
+		memcpy(ctx->pub, pub, 32);
+		ctx->pub_init = 1;
+	}
+
+	return 0;
+}
+
+int ica_x448_key_set(ICA_X448_CTX *ctx,
+		     const unsigned char priv[56],
+		     const unsigned char pub[56])
+{
+	if (check_fips() || !msa9_switch || ctx == NULL)
+		return -1;
+
+	if (priv != NULL) {
+		memcpy(ctx->priv, priv, 56);
+		ctx->priv_init = 1;
+		ctx->pub_init = 0;
+	}
+
+	if (pub != NULL) {
+		memcpy(ctx->pub, pub, 56);
+		ctx->pub_init = 1;
+	}
+
+	return 0;
+}
+
+int ica_ed25519_key_set(ICA_ED25519_CTX *ctx,
+			const unsigned char priv[32],
+			const unsigned char pub[32])
+{
+	if (check_fips() || !msa9_switch || ctx == NULL)
+		return -1;
+
+	if (priv != NULL) {
+		memcpy(ctx->sign_param.priv, priv, 32);
+		ctx->priv_init = 1;
+		ctx->pub_init = 0;
+	}
+
+	if (pub != NULL) {
+		s390_flip_endian_32(ctx->verify_param.pub, pub);
+		ctx->pub_init = 1;
+	}
+
+	return 0;
+}
+
+int ica_ed448_key_set(ICA_ED448_CTX *ctx,
+		      const unsigned char priv[57],
+		      const unsigned char pub[57])
+{
+	if (check_fips() || !msa9_switch || ctx == NULL)
+		return -1;
+
+	if (priv != NULL) {
+		memset(ctx->sign_param.priv, 0, sizeof(ctx->sign_param.priv));
+		memcpy(ctx->sign_param.priv + 64 - 57, priv, 57);
+		ctx->priv_init = 1;
+		ctx->pub_init = 0;
+	}
+
+	if (pub != NULL) {
+		memset(ctx->verify_param.pub, 0,
+                       sizeof(ctx->verify_param.pub));
+		memcpy(ctx->verify_param.pub, pub, 57);
+		s390_flip_endian_64(ctx->verify_param.pub,
+                                    ctx->verify_param.pub);
+		ctx->pub_init = 1;
+	}
+
+	return 0;
+}
+
+int ica_x25519_key_get(ICA_X25519_CTX *ctx, unsigned char priv[32],
+		       unsigned char pub[32])
+{
+	int rc;
+
+	if (check_fips() || !msa9_switch || ctx == NULL)
+		return -1;
+
+	if (priv != NULL) {
+		if (!ctx->priv_init)
+			return -1;
+
+		memcpy(priv, ctx->priv, 32);
+	}
+
+	if (pub != NULL) {
+		if (!ctx->pub_init) {
+			if (!ctx->priv_init)
+				return -1;
+
+			rc = x25519_derive_pub(ctx->pub, ctx->priv);
+			if (rc)
+				return -1;
+
+			ctx->pub_init = 1;
+		}
+
+		memcpy(pub, ctx->pub, 32);
+	}
+
+	return 0;
+}
+
+int ica_x448_key_get(ICA_X448_CTX *ctx, unsigned char priv[56],
+		     unsigned char pub[56])
+{
+	int rc;
+
+	if (check_fips() || !msa9_switch || ctx == NULL)
+		return -1;
+
+	if (priv != NULL) {
+		if (!ctx->priv_init)
+			return -1;
+
+		memcpy(priv, ctx->priv, 56);
+	}
+
+	if (pub != NULL) {
+		if (!ctx->pub_init) {
+			if (!ctx->priv_init)
+				return -1;
+
+			rc = x448_derive_pub(ctx->pub, ctx->priv);
+			if (rc)
+				return -1;
+
+			ctx->pub_init = 1;
+		}
+
+		memcpy(pub, ctx->pub, 56);
+	}
+
+	return 0;
+}
+
+int ica_ed25519_key_get(ICA_ED25519_CTX *ctx, unsigned char priv[32],
+			unsigned char pub[32])
+{
+	int rc;
+
+	if (check_fips() || !msa9_switch || ctx == NULL)
+		return -1;
+
+	if (priv != NULL) {
+		if (!ctx->priv_init)
+			return -1;
+
+		memcpy(priv, ctx->sign_param.priv,
+		       sizeof(ctx->sign_param.priv));
+	}
+
+	if (pub != NULL) {
+		if (!ctx->pub_init) {
+			if (!ctx->priv_init)
+				return -1;
+
+			rc = ed25519_derive_pub(ctx->verify_param.pub,
+						ctx->sign_param.priv);
+			if (rc)
+				return -1;
+
+			ctx->pub_init = 1;
+		}
+
+		s390_flip_endian_32(pub, ctx->verify_param.pub);
+	}
+
+	return 0;
+}
+
+int ica_ed448_key_get(ICA_ED448_CTX *ctx, unsigned char priv[57],
+			unsigned char pub[57])
+{
+	unsigned char pub64[64];
+	int rc;
+
+	if (check_fips() || !msa9_switch || ctx == NULL)
+		return -1;
+
+	if (priv != NULL) {
+		if (!ctx->priv_init)
+			return -1;
+
+		memcpy(priv, ctx->sign_param.priv + 64 - 57,
+		       sizeof(ctx->sign_param.priv) - (64 - 57));
+	}
+
+	if (pub != NULL) {
+		if (!ctx->pub_init) {
+			if (!ctx->priv_init)
+				return -1;
+
+			rc = ed448_derive_pub(ctx->verify_param.pub + 64 - 57,
+					      ctx->sign_param.priv + 64 - 57);
+			if (rc)
+				return -1;
+
+			ctx->pub_init = 1;
+		}
+
+		s390_flip_endian_64(pub64, ctx->verify_param.pub);
+		memcpy(pub, pub64, 57);
+	}
+
+	return 0;
+}
+
+int ica_x25519_derive(ICA_X25519_CTX *ctx,
+		      unsigned char shared_secret[32],
+		      const unsigned char peer_pub[32])
+{
+	int rc;
+
+	if (check_fips() || !msa9_switch || ctx == NULL
+	    || !ctx->priv_init || shared_secret == NULL || peer_pub == NULL)
+		return -1;
+
+	rc = scalar_mulx_cpacf(shared_secret, ctx->priv, peer_pub,
+			       NID_X25519);
+
+	stats_increment(ICA_STATS_X25519_DERIVE, ALGO_HW, ENCRYPT);
+	return rc;
+}
+
+int ica_x448_derive(ICA_X448_CTX *ctx,
+		    unsigned char shared_secret[56],
+		    const unsigned char peer_pub[56])
+{
+	int rc;
+
+	if (check_fips() || !msa9_switch || ctx == NULL
+	    || !ctx->priv_init || shared_secret == NULL || peer_pub == NULL)
+		return -1;
+
+	rc = scalar_mulx_cpacf(shared_secret, ctx->priv, peer_pub, NID_X448);
+
+	stats_increment(ICA_STATS_X448_DERIVE, ALGO_HW, ENCRYPT);
+	return rc;
+}
+
+int ica_ed25519_sign(ICA_ED25519_CTX *ctx, unsigned char sig[64],
+		     const unsigned char *msg, size_t msglen)
+{
+	int rc;
+
+	if (check_fips() || !msa9_switch || ctx == NULL
+	    || !ctx->priv_init || sig == NULL || (msg == NULL && msglen != 0))
+		return -1;
+
+	rc = s390_kdsa(S390_CRYPTO_EDDSA_SIGN_ED25519,
+		       &ctx->sign_param, msg, msglen);
+	if (rc)
+		return -1;
+
+	s390_flip_endian_32(sig, ctx->sign_param.sig);
+	s390_flip_endian_32(sig + 32, ctx->sign_param.sig + 32);
+	memset(ctx->sign_param.sig, 0, sizeof(ctx->sign_param.sig));
+
+	stats_increment(ICA_STATS_ED25519_SIGN, ALGO_HW, ENCRYPT);
+	return 0;
+}
+
+int ica_ed448_sign(ICA_ED448_CTX *ctx, unsigned char sig[114],
+		     const unsigned char *msg, size_t msglen)
+{
+	int rc;
+
+	if (check_fips() || !msa9_switch || ctx == NULL
+	    || !ctx->priv_init || sig == NULL || (msg == NULL && msglen != 0))
+		return -1;
+
+	rc = s390_kdsa(S390_CRYPTO_EDDSA_SIGN_ED448,
+		       &ctx->sign_param, msg, msglen);
+	if (rc)
+		return -1;
+
+	s390_flip_endian_64(ctx->sign_param.sig, ctx->sign_param.sig);
+	s390_flip_endian_64(ctx->sign_param.sig + 64,
+			    ctx->sign_param.sig + 64);
+	memcpy(sig, ctx->sign_param.sig, 57);
+	memcpy(sig + 57, ctx->sign_param.sig + 64, 57);
+	memset(ctx->sign_param.sig, 0, sizeof(ctx->sign_param.sig));
+
+	stats_increment(ICA_STATS_ED448_SIGN, ALGO_HW, ENCRYPT);
+	return 0;
+}
+
+int ica_ed25519_verify(ICA_ED25519_CTX *ctx, const unsigned char sig[64],
+		       const unsigned char *msg, size_t msglen)
+{
+	int rc;
+
+	if (check_fips() || !msa9_switch || ctx == NULL || sig == NULL
+	    || (msg == NULL && msglen != 0))
+		return -1;
+
+	if (!ctx->pub_init) {
+		if (!ctx->priv_init)
+			return -1;
+
+		rc = ed25519_derive_pub(ctx->verify_param.pub,
+				        ctx->sign_param.priv);
+		if (rc)
+			return -1;
+
+		ctx->pub_init = 1;
+	}
+
+	s390_flip_endian_32(ctx->verify_param.sig, sig);
+	s390_flip_endian_32(ctx->verify_param.sig + 32, sig + 32);
+
+	rc = s390_kdsa(S390_CRYPTO_EDDSA_VERIFY_ED25519,
+		       &ctx->verify_param, msg, msglen);
+	if (rc)
+		return -1;
+
+	memset(ctx->verify_param.sig, 0, sizeof(ctx->verify_param.sig));
+
+	stats_increment(ICA_STATS_ED25519_VERIFY, ALGO_HW, ENCRYPT);
+	return 0;
+}
+
+int ica_ed448_verify(ICA_ED448_CTX *ctx, const unsigned char sig[114],
+		     const unsigned char *msg, size_t msglen)
+{
+	int rc;
+
+	if (check_fips() || !msa9_switch || ctx == NULL || sig == NULL
+	    || (msg == NULL && msglen != 0))
+		return -1;
+
+	if (!ctx->pub_init) {
+		if (!ctx->priv_init)
+			return -1;
+
+		rc = ed448_derive_pub(ctx->verify_param.pub + 64 - 57,
+				      ctx->sign_param.priv + 64 - 57);
+		if (rc)
+			return -1;
+
+		ctx->pub_init = 1;
+	}
+
+	memcpy(ctx->verify_param.sig, sig, 57);
+	memcpy(ctx->verify_param.sig + 64, sig + 57, 57);
+	s390_flip_endian_64(ctx->verify_param.sig, ctx->verify_param.sig);
+	s390_flip_endian_64(ctx->verify_param.sig + 64,
+                            ctx->verify_param.sig + 64);
+
+	rc = s390_kdsa(S390_CRYPTO_EDDSA_VERIFY_ED448,
+		       &ctx->verify_param, msg, msglen);
+	if (rc || sig[113] != 0)	/* XXX kdsa doesnt check last byte */
+		return -1;
+
+	memset(ctx->verify_param.sig, 0, sizeof(ctx->verify_param.sig));
+
+	stats_increment(ICA_STATS_ED448_VERIFY, ALGO_HW, ENCRYPT);
+	return 0;
+}
+
+int ica_x25519_ctx_del(ICA_X25519_CTX **ctx)
+{
+	if (!msa9_switch || ctx == NULL || *ctx == NULL)
+		return -1;
+
+	OPENSSL_cleanse(*ctx, sizeof(**ctx));
+	free(*ctx);
+	*ctx = NULL;
+	return 0;
+}
+
+int ica_x448_ctx_del(ICA_X448_CTX **ctx)
+{
+	if (!msa9_switch || ctx == NULL || *ctx == NULL)
+		return -1;
+
+	OPENSSL_cleanse(*ctx, sizeof(**ctx));
+	free(*ctx);
+	*ctx = NULL;
+	return 0;
+}
+
+int ica_ed25519_ctx_del(ICA_ED25519_CTX **ctx)
+{
+	if (!msa9_switch || ctx == NULL || *ctx == NULL)
+		return -1;
+
+	OPENSSL_cleanse(*ctx, sizeof(**ctx));
+	free(*ctx);
+	*ctx = NULL;
+	return 0;
+}
+
+int ica_ed448_ctx_del(ICA_ED448_CTX **ctx)
+{
+	if (!msa9_switch || ctx == NULL || *ctx == NULL)
+		return -1;
+
+	OPENSSL_cleanse(*ctx, sizeof(**ctx));
+	free(*ctx);
+	*ctx = NULL;
+	return 0;
+}
+
+int ica_x25519_key_gen(ICA_X25519_CTX *ctx)
+{
+	if (check_fips() || !msa9_switch || ctx == NULL)
+		return -1;
+
+	memset(ctx, 0, sizeof(*ctx));
+
+	rng_gen(ctx->priv, 32);
+
+	ctx->priv_init = 1;
+	ctx->pub_init = 0;
+	return 0;
+}
+
+int ica_x448_key_gen(ICA_X448_CTX *ctx)
+{
+	if (check_fips() || !msa9_switch || ctx == NULL)
+		return -1;
+
+	memset(ctx, 0, sizeof(*ctx));
+
+	rng_gen(ctx->priv, 56);
+
+	ctx->priv_init = 1;
+	ctx->pub_init = 0;
+	return 0;
+}
+
+int ica_ed25519_key_gen(ICA_ED25519_CTX *ctx)
+{
+	if (check_fips() || !msa9_switch || ctx == NULL)
+		return -1;
+
+	memset(ctx, 0, sizeof(*ctx));
+
+	rng_gen(ctx->sign_param.priv, sizeof(ctx->sign_param.priv));
+
+	ctx->priv_init = 1;
+	ctx->pub_init = 0;
+	return 0;
+}
+
+int ica_ed448_key_gen(ICA_ED448_CTX *ctx)
+{
+	if (check_fips() || !msa9_switch || ctx == NULL)
+		return -1;
+
+	memset(ctx, 0, sizeof(*ctx));
+
+	rng_gen(ctx->sign_param.priv + 64 - 57,
+		sizeof(ctx->sign_param.priv) - (64 - 57));
+
+	ctx->priv_init = 1;
+	ctx->pub_init = 0;
+	return 0;
+}
+
+
 /*
  *                             End of ECC API
  *
